#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kxy.api.core import least_mixed_mutual_information, discrete_mutual_information

def classification_suboptimality(yp, y, x_c, x_d=None):
	"""
	.. _classification-suboptimality:
	Quantifies the extent to which a (multinomial) classifier can be improved without requiring additional features.

	.. note::

		The conditional entropy :math:`h \\left( y \\vert x \\right)` represents the amount of information 
		about :math:`y` that cannot be explained by :math:`x`. If we denote :math:`f(x)` the label 
		predicted by our classifier, :math:`h \\left( y \\vert f(x) \\right)` represents the amount
		of information about :math:`y` that the classifier is not able to explain using :math:`x`.

		A natural metric for how suboptimal a particular classifier is can therefore be defined as the 
		difference between the amount of information about :math:`y` that cannot be explained by 
		:math:`f(x)` and the amount of information about :math:`y` that cannot be explained by :math:`x`

		.. math::

			\\text{SO}(f; x) &= h \\left( y \\vert f(x) \\right) - h \\left( y \\vert x \\right) \\

										 :&= I\\left(y, x \\right) - I\\left(y, f(x) \\right) \\

										  &\\geq 0.

		This classification suboptimality metric is 0 if and only if :math:`f(x)` fully captures any information about :math:`y`
		that is contained in :math:`x`. When 

		.. math::

			\\text{SO}(f; x) > 0 

		on the other hand, there exists a classification model using :math:`x` as features that can better predict :math:`y`. The larger 
		:math:`\\text{SO}(f; x)`, the more the classification model is suboptimal and can be improved.


	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous features generating distribution.
	x_d : (n,) np.array
		n i.i.d. draws from the discrete features generating distribution, if any, sampled jointly with x_c.
	y : (n,) np.array
		n i.i.d. draws from the (discrete) labels generating distribution, sampled
		jointly with x_c and x_d.
	yp : (n,) np.array
		Discrete predictions of y from our classifier.


	Returns
	-------
	d : float
		The classifier's suboptimality measure.


	.. seealso:: 

		:ref:`kxy.api.core.mutual_information.least_mixed_mutual_information <least-mixed-mutual-information>`
	"""
	return least_mixed_mutual_information(x_c, y, x_d=x_d)-discrete_mutual_information(y, yp)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Extension of pandas DataFrame class to allow data scientists to tap into 
the power of the KXY API within the comfort of their favorite data structure.
"""

from functools import lru_cache, wraps
from threading import Thread

import numpy as np
import pandas as pd

from kxy.api.core import least_total_correlation, spearman_corr
from kxy.api import solve_copula_async
from kxy.classification import classification_feasibility, classification_suboptimality
from kxy.finance import information_adjusted_beta, information_adjusted_correlation
from kxy.regression import regression_feasibility, regression_suboptimality, regression_additive_suboptimality

from .decorators import decorate_methods, decorate_all_methods


def cast_to_kxy_dataframe(method):
	"""
	Decorator casting the return of a method to kxy.DataFrame when it returns pd.DataFrame.
	"""
	@wraps(method)
	def wrapper(*args, **kwargs):
		res = method(*args, **kwargs)
		if isinstance(res, pd.DataFrame):
			return DataFrame(data=res.values, columns=res.columns, index=res.index)
		return res

	return wrapper


@decorate_all_methods(cast_to_kxy_dataframe)
class _iLocIndexer(pd.core.indexing._iLocIndexer):
	pass

@decorate_all_methods(cast_to_kxy_dataframe)
class _LocIndexer(pd.core.indexing._LocIndexer):
	pass

@decorate_all_methods(cast_to_kxy_dataframe)
class _iAtIndexer(pd.core.indexing._iAtIndexer):
	pass

@decorate_all_methods(cast_to_kxy_dataframe)
class _AtIndexer(pd.core.indexing._AtIndexer):
	pass

@decorate_methods(cast_to_kxy_dataframe, include=[\
	'__add__', '__sub__', '__mul__', '__div__', '__floordiv__', '__truediv__', '__mod__', '__pow__', \
	'__iadd__', '__isub__', '__imul__', '__idiv__', '__ifloordiv__', '__imod__', '__ipow__', \
	'add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'sample', \
	'radd', 'rsub', 'rmul', 'rdiv', 'rtruediv', 'rfloordiv', 'rmod', 'rpow', \
	'__eq__', '__ne__', '__gt__', '__lt__', '__le__', '__ge__', '__neg__', '__pos__',  '__invert__', \
	'eq', 'ne', 'gt', 'lt', 'le', 'ge', 'apply', 'applymap', 'abs', 'clip', 'count', \
	'cov', 'cummin', 'cummax', 'cumprod', 'cumsum', 'round', 'sum', 'std', 'var', 'cov', \
	'nunique', 'rename', 'rename_axis', 'reset_index', 'drop_level', 'pivot', 'T', 'rank', \
	'transpose', 'fillna', 'dropna', 'pct_change', 'swapaxes', 'swaplevel', 'where', 'tail', \
	'head', 'shift', 'asfreq', 'dot', 'transform', 'astype', 'copy'])
class DataFrame(pd.DataFrame):
	"""
	Extension of the pandas.DataFrame class with various analytics for **pre-learning** and **post-learning**,
	in supervised learning problems.

	.. important::

		A wide range of operators and methods inherited from pandas.DataFrame have been overriden to return a kxy.DataFrame
		instead of a pandas.DataFrame.
	"""
	def regression_feasibility(self, label_column, input_columns=()):
		"""
		Quantifies how feasible a regression problem is by computing the amount of uncertainty
		about the label that can be reduced by knowing the inputs.

		.. math::
			FE(x; y) &= h(y)-h\\left(y \\vert x \\right) \\

									  &= I(y, x)

									  &= h\\left(u_x\\right)-h\\left(u_{x,y}\\right)

		Copula entropies are estimated by solving a maximum-entropy copula problem under concordance-like
		constraints on x and on (y, x) jointly. 


		.. seealso::

			:ref:`kxy.regression.pre_leanring.regression-feasibility <regression-feasibility>`.


		Parameters
		----------
		label_column : str
			The name of the column to use as label.
		input_columns : set, optional
			The set of columns to as inputs. When input_columns is the empty set,
			all columns except for label_column are used as inputs.

		Returns
		-------
		f : float
			The feasibility score in :math:`[0, \\infty]`. The larger the better.

		Raises
		------
		AssertionError
			If label_column is in input_columns.
		"""
		input_columns = list(set(self.columns)-set([label_column])) if input_columns == () \
			else list(input_columns)
		assert label_column not in input_columns, "The output cannot be a input"

		input_importance = self.input_importance(label_column, input_columns=input_columns,\
			problem='regression')
		input_sorted_by_importance = input_importance['input'].values

		return regression_feasibility(self[input_sorted_by_importance].values, self[label_column].values)


	def classification_feasibility(self, label_column, discrete_input_columns=(), \
			continuous_input_columns=()):
		"""
		Quantifies how feasible a classification problem is by computing the amount of uncertainty
		about the label that can be reduced by knowing the inputs, in a model-free fashion.

		.. math::
			FE(x; y) &= h(y)-h\\left(y \\vert x \\right) \\

									 &= I(y, x)

		.. seealso::

			:ref:`kxy.classification.pre_leanring.classification_feasibility <classification-feasibility>`.

		Parameters
		----------
		label_column : str
			The name of the column to use as label.
		discrete_input_columns : set, optional
			The set of columns, if any, to use as inputs that are discrete.
		continuous_input_columns : set, optional
			The set of columns to use as inputs that are continuous.

		Returns
		-------
		f : float
			The feasibility score in :math:`[0, \\infty]`. The larger the better.
		"""
		assert label_column not in discrete_input_columns, "The output cannot be a input"
		assert label_column not in continuous_input_columns, "The output cannot be a input"

		continuous_input_columns = [col for col in self.columns if col != label_column and not self.is_discrete(col)] \
			if continuous_input_columns == () else list(continuous_input_columns)
		assert len(continuous_input_columns) > 0, "Continuous inputs are required"

		x_d = self[discrete_input_columns].values if len(discrete_input_columns) > 0 else None
		x_c = self[continuous_input_columns].values
		y = self[label_column].values

		return classification_feasibility(x_c, y, x_d=x_d)



	def classification_suboptimality(self, prediction_column, label_column, discrete_input_columns=(), \
			continuous_input_columns=()):
		"""
		Quantifies the extent to which a (multinomial) classifier can be improved without requiring additional inputs.

		.. note::

			The conditional entropy :math:`h \\left( y \\vert x \\right)` represents the amount of information 
			about :math:`y` that cannot be explained by :math:`x`. If we denote :math:`f(x)` the label 
			predicted by our classifier, :math:`h \\left( y \\vert f(x) \\right)` represents the amount
			of information about :math:`y` that the classifier is not able to explain using :math:`x`.

			A natural metric for how suboptimal a particular classifier is can therefore be defined as the 
			difference between the amount of information about :math:`y` that cannot be explained by 
			:math:`f(x)` and the amount of information about :math:`y` that cannot be explained by :math:`x`

			.. math::

				\\text{SO}(f; x) &= h \\left( y \\vert f(x) \\right) - h \\left( y \\vert x \\right) \\

				:&= I\\left(y, x \\right) - I\\left(y, f(x) \\right) \\

				 &\\geq 0.

			This classification suboptimality metric is 0 if and only if :math:`f(x)` fully captures any information about :math:`y`
			that is contained in :math:`x`. When 

			.. math::

				\\text{SO}(f; x) > 0 

			on the other hand, there exists a classification model using :math:`x` as inputs that can better predict :math:`y`. The larger 
			:math:`\\text{SO}(f; x)`, the more the classification model is suboptimal and can be improved.


		Parameters
		----------
		prediction_column : str
			The column containing predicted labels.
		label_column : str
			The column containing true labels.
		continuous_input_columns : set
			The set of columns containing continuous inputs.
		discrete_input_columns : set
			The set of columns containing discrete input, if any.


		Returns
		-------
		d : float
			The classifier's suboptimality measure.


		.. seealso::

			:ref:`kxy.classification.post_learning.classification_suboptimality <classification-suboptimality>`
		"""
		assert label_column not in discrete_input_columns, "The output cannot be a input"
		assert label_column not in continuous_input_columns, "The output cannot be a input"

		continuous_input_columns = [col for col in self.columns if col not in (label_column, prediction_column) \
			and not self.is_discrete(col)] if continuous_input_columns == () else list(continuous_input_columns)
		assert len(continuous_input_columns) > 0, "Continuous inputs are required"

		x_d = self[discrete_input_columns].values if len(discrete_input_columns) > 0 else None
		x_c = self[continuous_input_columns].values
		y = self[label_column].values
		yp = self[prediction_column].values

		return classification_suboptimality(yp, y, x_c, x_d=x_d)



	def input_importance(self, label_column, input_columns=(), problem=None):
		"""
		Calculates the importance of each input in the input set at solving the supervised
		learning problem where the label is defined by the label_column.

		
		.. note::

			Input importance is defined as the mutual information between the input column 
			and the label column. The supervised learning problem can either be specified as :code:`'classification'` or 
			:code:`'regression'` using the :code:`problem` argument, or inferred from the type of, and the number 
			of distinct values in the :code:`label_column`.


		.. seealso::

			* :ref:`kxy.classification.pre_leanring.classification_feasibility <classification-feasibility>`
			* :ref:`kxy.classification.pre_leanring.regression_feasibility <regression-feasibility>`


		Parameters
		----------
		label_column : str
			The name of the column to use as label.
		input_columns : set, optional
			The set of columns to as inputs. When input_columns is the empty set,
			all columns except for label_column are used as inputs.
		problem : str or None (default), optional
			The type of supervised learning problem. One of None (default), 'classification'
			or 'regression'. When problem is None, the supervised learning problem is inferred
			based on whether labels are numeric and the percentage of distinct labels.

		Returns
		-------
		importance : DataFrame
			A dataframe with as input column, an importance column, and a normalized importance column.

		Raises
		------
		AssertionError
			If problem is neither :code:`None` nor :code:`'classification'` nor :code:`'regression'`, or if 
			:code:`label_column` is in :code:`input_columns`.
		"""
		input_columns = list(set(self.columns)-set([label_column])) if input_columns == () \
			else list(input_columns)
		assert label_column not in input_columns, "The output cannot be a input"
		assert problem is None or problem in ('classification', 'regression'), \
			"The problem should be either None, 'classification' or 'regression'"

		if problem is None:
			problem = 'classification' if self.is_discrete(label_column) else 'regression'

		if problem == 'classification':
			importance = {col: classification_feasibility(\
							None, self[label_column].values, x_d=self[col].values) if self.is_discrete(col) else \
							   classification_feasibility(self[col].values, self[label_column].values, x_d=None) \
							for col in input_columns}

		else:
			importance = {col: regression_feasibility(\
				self[col].values, self[label_column].values) if not self.is_categorical(col) else \
				classification_feasibility(self[label_column].values, self[col].values, x_d=None) \
				for col in input_columns}

		total_importance = np.sum([importance[col] for col in importance.keys() if importance[col]])
		scale = 1./total_importance if total_importance > 0. else 0.0

		importance_df = DataFrame({
			'input': [k for k, v in sorted(importance.items(), key=lambda item: -item[1])], \
			'importance': [v for k, v in sorted(importance.items(), key=lambda item: -item[1])], \
			'normalized_importance': [v*scale for k, v in sorted(importance.items(), key=lambda item: -item[1])]})

		return importance_df


	def is_discrete(self, column):
		"""
		Determine whether the input column contains discrete observations.
		"""
		ret = (not np.can_cast(self[column].values, float))
		ret = ret or len(list(set(self[column].values))) < 0.5*self.shape[0]

		return ret


	def is_categorical(self, column):
		"""
		Determinee whether the input column is not numeric.
		"""
		return not np.can_cast(self[column].values, float)


	def corr(self, columns=(), method='information-adjusted', min_periods=1, p=0):
		"""
		Calculates the auto-correlation matrix of all columns or the input subset.


		Parameters
		----------
		columns : set, optional
			The set of columns to use. If not provided, all columns are used.
		method : str, optional
			Which method to use to calculate the auto-correlation matrix. Supported
			values are 'information-adjusted' (the default) and all 'method' values of pandas.DataFrame.corr.
		p : int, optional
			The number of lags to use when generating Spearman rank auto-correlation to use 
			as empirical evidence in the maximum-entropy problem. The default value is 0, which 
			corresponds to assuming rows are i.i.d. This is also the only supported value for now.
		min_periods : int, optional
			Only used when method is not 'information-adjusted'. 
			See the documentation of pandas.DataFrame.corr.

		Returns
		-------
		c : DataFrame
			The auto-correlation matrix.


		.. seealso::

			:ref:`kxy.finance.risk_analysis.information_adjusted_correlation <information-adjusted-correlation>`
		"""
		columns = self.columns if columns == () else list(columns)

		if method == 'information-adjusted':
			c = information_adjusted_correlation(self[columns].values, self[columns].values)
			return pd.DataFrame(c, columns=columns, index=columns)
		else:
			return pd.DataFrame.corr(self[columns], method=method, min_periods=min_periods)


	def beta(self, column_y, column_x, method='information-adjusted'):
		"""
		Calculates the information-adjusted beta of a portfolio or asset (whose returns are provided in 
		column_y) with respect to the market (whose returns are provided in column_x).

		.. note::

			The information-adjusted beta coefficient generalizes the traditional (CAPM/OLS/Pearson) beta 
			in that, unlike CAPM beta that only captures linear cross-sectional dependence, the 
			information-adjusted beta captures cross-sectional and temporal dependence, linear and nonlinear.

			The IA-beta is 0 if and only if the portfolio or asset exhibit no dependence with the market, linear
			or nonlinear, cross-sectional or temporal.


		Parameters
		----------
		colummn_y : str
			The name of the column to use for portfolio/asset returns.
		colummn_x : str
			The name of the column to use for market returns.
		method : str, optional
			Either 'information-adjusted' for information-adjusted beta or 'pearson' for the traditional OLS/pearson beta coefficient.

		Returns
		-------
		c : float
			The beta coefficient.

		Raises
		------
		AssertionError
			If the method is neither 'information-adjusted' nor 'pearson'.


		.. seealso::

			:ref:`kxy.finance.factor_analysis.information_adjusted_beta <information-adjusted-beta>`
		"""
		assert method in ('information-adjusted', 'pearson'), "Allowed methods are 'information-adjusted' and 'pearson'."

		if method == 'information-adjusted':
			return information_adjusted_beta(self[column_y].values, self[column_x].values)

		c = np.corrcoef(self[column_y].values, self[column_x].values)[0, 1]
		return c*np.sqrt(self[column_y].values.var()/self[column_x].values.var())


	def total_correlation(self, columns=()):
		"""
		Calculates the total correlation between all columns or the input subset.


		Parameters
		----------
		columns : set, optional
			The set of columns to use. If not provided, all columns are used.

		Returns
		-------
		c : float
			The total correlation.


		.. seealso::

			:ref:`kxy.api.core.mutual_information.least_total_correlation <least-total-correlation>`
		"""
		columns = self.columns if columns == () else list(columns)
		return least_total_correlation(self[columns].values)


	def regression_suboptimality(self, prediction_column, label_column, input_columns=()):
		"""
		Quantifies the extent to which a calibrated regression model can be improved without requiring
		additional inputs.


		.. note::

			The aim of a regression model is to find the function :math:`x \\to f(x) \in \\mathbb{R}` 
			to be used as our predictor for :math:`y` given :math:`x` so that :math:`f(x)` fully captures all
			the information about :math:`y` that is contained in :math:`x`.  For instance, this will be 
			the case when the true generative model takes the form

			.. math::

				y = f(x) + \\epsilon

			with :math:`\\epsilon` statistically independent from :math:`y`, in which case :math:`h \\left( y \\vert x \\right) = h(\\epsilon)`.

			More generally, the conditional entropy :math:`h \\left( y \\vert x \\right)` represents 
			the amount of information about :math:`y` that cannot be explained by :math:`x`, while 
			:math:`h \\left( y \\vert f(x) \\right)` represents the amount of information 
			about :math:`y` that cannot be explained by the regression model 

			.. math::

				y = f(x) + \\epsilon.

			A natural metric for how suboptimal a particular regression model is can therefore be defined as
			the difference between what the amount of information about :math:`y` that cannot be explained by 
			:math:`f(x)` and the amount of information about :math:`y` that cannot be explained by :math:`x`


			.. math::

				\\text{SO}(f; x) &= h \\left( y \\vert f(x) \\right) - h \\left( y \\vert x \\right) \\

				:&= I\\left(y, x \\right) - I\\left(y, f(x) \\right) \\

				 &\\geq 0.

			This regression suboptimality metric is 0 if and only if :math:`f(x)` fully captures any information about :math:`y`
			that is contained in :math:`x`. When 

			.. math::

				\\text{SO}(f; x) > 0 

			on the other hand, there exists a regression model using :math:`x` as inputs that can better predict :math:`y`. The larger 
			:math:`\\text{SO}(f; x)`, the more the regression model is suboptimal and can be improved.


		Parameters
		----------
		prediction_column : str
			The name of the column containing regression predictions.
		label_column : str
			The name of the column containing regression labels.
		input_columns: set
			The set of columns to use as inputs. When not specified, all columns 
			are used except. for the prediction and label columns.

		Returns
		-------
		s : float
			The suboptimality score, defined as the mutual information between residuals and labels.
			The higher the value, the more the model can be improved.


		.. seealso::

			:ref:`kxy.regression.post_learning.regression_suboptimality <regression-suboptimality>`
		"""
		assert not self.is_categorical(prediction_column), "The prediction column should not be categorical"
		assert not self.is_categorical(label_column), "The label column should not be categorical"

		input_columns = [_ for _ in self.columns if _ not in (prediction_column, label_column) ] \
			if input_columns == () else list(input_columns)

		return regression_suboptimality(self[prediction_column].values, self[label_column].values, \
			self[input_columns].values)




	def regression_additive_suboptimality(self, prediction_column, label_column, input_columns=()):
		"""
		Quantifies the extent to which a regression model can be improved without requiring additional inputs, by evaluating 
		how informative its residuals still are about the inputs.

		.. note::

			Additive regression models aim at breaking down a label :math:`y` as the sum of a component that solely depend on 
			inputs :math:`f(x)` and a residual component that is statistically independent from inputs :math:`\\epsilon`

			.. math::

				y = f(x) + \\epsilon.

			In an ideal scenario, the regreession residual :math:`\\epsilon` would indeed be stastically independent from the inputs
			:math:`x`. In pratice however, this might not be the case, for instance when the space of candidate functions used by
			the regression model isn't flexible enough (e.g. linear regression or basis functions regression), or the optimization
			has not converged to the global optimum. 

			Any departure from statistical independence between residuals :math:`\\epsilon` and inputs :math:`x` is an indication that what
			:math:`x` can reveal about :math:`y` is not fully captured by :math:`f(x)`, which implies that the regression model can be improved.

			Thus, we define the additive suboptimality of a regression model as the mutual information between its residuals and its inputs

			.. math::

				\\text{ASO}(f; x) := I\\left( y-f(x), x \\right)


		Parameters
		----------
		prediction_column : str
			The name of the column containing regression predictions.
		label_column : str
			The name of the column containing regression labels.
		input_columns: set
			The set of columns to use as inputs. When not specified, all columns 
			are used except. for the prediction and label columns.


		Returns
		-------
		d : float
			The regression's additive suboptimality measure.


		.. seealso::

			:ref:`kxy.regression.post_learning.regression_additive_suboptimality <regression-additive-suboptimality>`
		"""
		assert not self.is_categorical(prediction_column), "The prediction column should not be categorical"
		assert not self.is_categorical(label_column), "The label column should not be categorical"

		input_columns = [_ for _ in self.columns if _ not in (prediction_column, label_column) ] \
			if input_columns == () else list(input_columns)

		e = (self[prediction_column]-self[label_column]).values
		x = self[input_columns].values

		return regression_additive_suboptimality(e, x)


	def __hash__(self):
		return hash(self.to_string())


	@property
	def iloc(self):
		"""
		Ensures the inherited iloc operations return a DataFrame instead of a pandas.DataFrame.
		"""
		return _iLocIndexer("iloc", self)


	@property
	def loc(self):
		"""
		Ensures the inherited loc operations return a DataFrame instead of a pandas.DataFrame.
		"""
		return _LocIndexer("loc", self)


	@property
	def at(self):
		"""
		Ensures the inherited at operations return a DataFrame instead of a pandas.DataFrame.
		"""
		return _AtIndexer("at", self)


	@property
	def iat(self):
		"""
		Ensures the inherited iat operations return a DataFrame instead of a pandas.DataFrame.
		"""
		return _iAtIndexer("iat", self)


	def _pre_solve(self):
		"""
		Pre-emptively solve the max-entropy problems remotely, and in the background.
		"""
		data = np.hstack((self.values, np.abs(self.values-self.values.mean(axis=0))))
		corr = spearman_corr(data)
		return solve_copula_async(corr)
		
	



@cast_to_kxy_dataframe
def read_csv(*args, **kwargs):
	"""
	.. _read-csv:
	Same as pandas.read_csv, but returns a kxy.DataFrame object.
	"""
	return pd.read_csv(*args, **kwargs)


@cast_to_kxy_dataframe
def read_excel(*args, **kwargs):
	"""
	.. _read-excel:
	Same as pandas.read_excel, but returns a kxy.DataFrame object.
	"""
	return pd.read_excel(*args, **kwargs)


@cast_to_kxy_dataframe
def read_html(*args, **kwargs):
	"""
	.. _read-html:
	Same as pandas.read_html, but returns a kxy.DataFrame object.
	"""
	return pd.read_html(*args, **kwargs)


@cast_to_kxy_dataframe
def read_table(*args, **kwargs):
	"""
	.. _read-table:
	Same as pandas.read_table, but returns a kxy.DataFrame object.
	"""
	return pd.read_table(*args, **kwargs)


@cast_to_kxy_dataframe
def read_sql(*args, **kwargs):
	"""
	.. _read-sql:
	Same as pandas.read_sql, but returns a kxy.DataFrame object.
	"""
	return pd.read_sql(*args, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import requests
from time import sleep

from .client import APIClient



def solve_copula_async(corr):
	"""
	.. _solve-copula-async:
	Solve the maximum-entropy copula problem under Spearman rank correlation matrix constraints asynchronously.

	.. note:: 

		The solution to the optimization problem is not returned. This function should be used to pre-compute the solution 
		to the optimization problem for later use.


	.. seealso::

		:ref:`kxy.api.optimizers.solve_sync <solve-copula-sync>`.


	Parameters
	----------
	corrr : np.array
		The Spearman correlation matrix.


	Returns
	-------
	success : bool
		Whether the request went through successfully.
	"""
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	logging.debug('Launching a max-ent solver in the background with corr=%s' % c)
	api_response = APIClient.route(path='/core/dependence/copula/maximum-entropy/entropy/rv/pre-compute', \
		method='GET', corr=c)

	if api_response.status_code == requests.codes.ok:
		return True

	logging.info(api_response.json())
	return False




def solve_copula_sync(corr, mode=None, output_index=None, input_indices=None, condition_indices=None):
	"""
	.. _solve-copula-sync:
	Solve the maximum-entropy copula problem under Spearman rank correlation matrix constraints synchronously.

	.. note:: 

		This function blocks until the optimization problem is solved, and returrns the requested quantity.

	.. seealso::

		:ref:`kxy.api.optimizers.solve_sync <solve-copula-async>`.


	Parameters
	----------
	corrr : np.array
		The Spearman correlation matrix.

	mode : str
		One of :code:`'copula_entropy'`, :code:`'mutual_information_v_output'`, and :code:`'conditional_mutual_information'`.

		When mode is :code:`'copula_entropy'` the function returns the entropy of the copula of the random
		vector whose Spearman correlation matrix is the input :code:`corr`.

		When mode is :code:`'mutual_information_v_output'` the function returns the mutual information between 
		a continuous output variable :math:`y` (specificied by :code:`output_index`) and continuous input 
		variables :math:`x` (specificied by the other variables): :math:`I(x, y)`. We recall that the mutual information 
		between two continuous variables is the mutual information of their copula-uniform dual representations. The copula 
		of :math:`(x, y)` is learned as the maximum-entropy copula under the constraint that the Spearman correlation matrix 
		is the input :code:`corr`.

		When mode is :code:`'conditional_mutual_information'` the function returns the conditional mutual information
		between output :math:`y` (specified by column :code:`output_index`) and inputs :math:`x` 
		(specified by columns :code:`input_indices`) conditional on :math:`z` (specified by columns :code:`condition_indices`):
		:math:`I(x, y\\vert z)`. We recall that the conditional mutual information of two random variables given a third, 
		is the conditional mutual information between the associated copula-uniform dual representations. The copula of 
		:math:`(x, y, z)` is learned as the maximum-entropy copula under the constraint that its Spearman correlation matrix 
		is given by the input :code:`corr`.

	output_index : int
		The index of the column that should be used as output variable when mode is either :code:`'mutual_information_v_output'` or
		:code:`'conditional_mutual_information'`.

	input_indices : list
		The indices of the columns that should be used as input variables when mode is :code:`'conditional_mutual_information'`.

	condition_indices : list
		The indices of the columns that should be used as condition variables when mode is :code:`'conditional_mutual_information'`.


	Returns
	-------
	r : float
		The requested result, namely the copula entropy, the mutual information or the conditional
		mutual information, depending on the value of the mode.
	"""
	assert mode in ('copula_entropy', 'mutual_information_v_output', 'conditional_mutual_information')

	if mode in ['mutual_information_v_output', 'conditional_mutual_information']:
		assert output_index is not None, 'The output index should be provided'

	if mode == 'conditional_mutual_information':
		assert input_indices is not None, 'Input indices should be provided for conditional mutual information.'
		assert condition_indices is not None, 'Condition indices should be provided for conditional mutual information.'

	# solve_copula_async(corr)
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	opt_launched = False
	max_retry = 60
	first_try = True
	retry_count = 0
	while (first_try or api_response.status_code == requests.codes.not_found) and retry_count < max_retry:
		logging.debug('Querying the max-ent solution with corr=%s and mode=%s' % (c, mode))
		first_try = False
		if mode == 'copula_entropy':
			api_response = APIClient.route(\
				path='/core/dependence/copula/maximum-entropy/entropy/rv/all', method='GET', \
				corr=c, mode='copula_entropy')

		if mode == 'mutual_information_v_output':
			api_response = APIClient.route(\
				path='/core/dependence/copula/maximum-entropy/entropy/rv/all', method='GET', \
				corr=c, mode='mutual_information_v_output', output_index=output_index)

		if mode == 'conditional_mutual_information':
			input_indices = json.dumps(input_indices)
			condition_indices = json.dumps(condition_indices)
			api_response = APIClient.route(\
				path='/core/dependence/copula/maximum-entropy/entropy/rv/all', method='GET', \
				corr=c, mode='conditional_mutual_information', output_index=output_index, \
				input_indices=input_indices, condition_indices=condition_indices)

		retry_count += 1
		if api_response.status_code == requests.codes.not_found:
			sleep(10.)

	logging.debug(api_response.json())
	if api_response.status_code == requests.codes.ok:
		if mode == 'copula_entropy':
			h = api_response.json()['entropy']
			if h is not None:
				return float(h)

		if mode == 'mutual_information_v_output':
			mi = api_response.json()['mutual_information']
			if not mi is None:
				return float(mi)

		if mode == 'conditional_mutual_information':
			mi = api_response.json()['conditional_mutual_information']
			if not mi is None:
				return float(mi)

	return None





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
==================
kxy.api.decorators
==================
"""

from functools import wraps
import os

API_KEY = os.getenv('KXY_API_KEY')

def requires_api_key(method):
	"""
	Decorator used to make functions and methods calls fail
	when they require an API key and the environment variable
	'KXY_API_KEY' is not set. The decorated function or method 
	is otherwise not affected.

	Raises
	------
	AssertionError
		If the environment variable 'KXY_API_KEY' is not set.
	"""
	@wraps(method)
	def wrapper(*args, **kw):
		assert API_KEY is not None, 'An API key should be provided by setting the environment' \
			'variable KXY_API_KEY'
		return method(*args, **kw)

	return wrapper


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import requests

import numpy as np
import scipy.special as spe

from kxy.api import APIClient, solve_copula_sync
from .utils import avg_pairwise_spearman_corr, pre_conditioner, spearman_corr


def scalar_continuous_entropy(x):
	"""
	.. _scalar-continuous-entropy:
	Estimates the (differential) entropy of a continuous scalar random variable using the standard 1-spacing estimator (see [2]_ and [3]_):

	.. math::
		h(x) \\approx - \gamma(1) + \\frac{1}{n-1} \\sum_{i=1}^{n-1} \log \\left[ n \\left(x_{(i+1)} - x_{(i)} \\right) \\right],

	where :math:`x_{(i)}` is the i-th smallest sample, and :math:`\\gamma` is
	the digamma function. 

	
	.. note::

		A random noise with negligible standard deviation is added to the input
		for robustness.

	
	Parameters
	----------
	x : (n,) np.array
		i.i.d. samples from the distribution of interest.

	Returns
	-------
	h : float
		The (differential) entropy of the continuous scalar random variable of interest.

	Raises
	------
	AssertionError
		If the input has the wrong shape.
		

	.. rubric:: References

	.. [2] Kozachenko, L. F., and Nikolai N. Leonenko. "Sample estimate of the entropy of a random vector." 
		Problemy Peredachi Informatsii 23.2 (1987): 9-16.

	.. [3] Beirlant, J., Dudewicz, E.J., Györfi, L., van der Meulen, E.C. "Nonparametric entropy estimation: an overview." 
		International Journal of Mathematical and Statistical Sciences. 6 (1): 17–40. (1997) ISSN 1055-7490. 
	"""
	assert len(x.shape) == 1 or x.shape[1] == 1, 'x should be a one dimensional numpy array'
	sx = np.unique(x)
	dx = np.min(sx[1:]-sx[:-1])/100.0
	eps = np.random.rand(*x.shape)*dx
	x += eps
	sorted_x = np.unique(x)
	n = sorted_x.shape[0]
	ent = np.sum(np.log(n*(sorted_x[1:]-sorted_x[:-1])))/n - spe.digamma(1)

	return ent


def discrete_entropy(x):
	"""
	.. _discrete-entropy:
	Estimates the (Shannon) entropy of a discrete random variable taking up to q distinct values 
	given n i.i.d samples.

	.. math::
		h(x) = - \sum_{i=1}^q p_i \log p_i

	Parameters
	----------
	x : (n,) np.array
		i.i.d. samples from the distribution of interest.

	Returns
	-------
	h : float
		The (differential) entropy of the discrete random variable of interest.

	Raises
	------
	AssertionError
		If the input has the wrong shape.

	"""
	assert len(x.shape) == 1, 'Only flat numpy arrays are supported'
	categories = list(set(list(x)))
	n = len(x)
	probas = np.array([1.*len(x[x==cat])/n for cat in categories])

	return -np.dot(np.log(probas), probas)


def least_structured_copula_entropy(x):
	"""
	.. _least-structured-copula-entropy:
	Estimates the entropy of the least informative copula model whose average pairwise
	Spearman rank correlation is the same as the sample estimate from the input
	array.

	.. note::
	
		This also corresponds to least amount of total correlation that is evidenced
		by the sample average pairwise Spearman's rank correlation.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	Returns
	-------
	h : float
		The (differential) entropy of the least structured copula 
		consistent with observed average pairwise Spearman's rank correlation. See also :ref:`max_ent_copula_entropy <max-ent-copula-entropy>`.
	"""
	if len(x.shape) == 1 or x.shape[1] == 1:
		# By convention, the copula-dual representation 
		# of a 1d random variable is the uniform[0, 1].
		return 0.0

	corr = spearman_corr(x)
	h = solve_copula_sync(corr, mode='copula_entropy')

	return h


def least_structured_continuous_entropy(x):
	""" 
	.. _least-structured-continuous-entropy:
	Estimates the entropy of a continuous d-dimensional random variable under the 
	least structured assumption for its copula. When :math:`d>1`,

	.. math::
		h(x) = h(u) + \sum_{i=1}^d h(x_i),

	and :math:`h(u)` is estimated using :ref:`least_structured_copula_entropy <least-structured-copula-entropy>` and 
	:math:`h(x_i)` are estimated using :ref:`scalar_continuous_entropy <scalar-continuous-entropy>`. 


	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	Returns
	-------
	h : float
		The (differential) entropy of the data generating distribution, assuming 
		its copula is the least structured copula consistent with the observed 
		average pairwise Spearman's rank correlation. 

		By convention, when :math:`d=1`, this function is the same as :ref:`scalar_continuous_entropy. <scalar-continuous-entropy>`, 
		and returns 0 when :math:`n=1`.
	"""
	if x.shape[0] == 1:
		# By convention, the entropy of a single sample is 0.
		return 0.0

	if len(x.shape) == 1 or x.shape[1] == 1:
		return scalar_continuous_entropy(x)

	x = x - x.mean(axis=0)
	a, log_abs_det_a = pre_conditioner(x.T)
	z = np.dot(a, x.T).T

	ch = least_structured_copula_entropy(z)
	ih = np.sum([scalar_continuous_entropy(z[:, i]) for i in range(z.shape[1])])

	return ih+ch-log_abs_det_a


def least_structured_mixed_entropy(x_c, x_d):
	"""
	.. _least-structured-mixed-entropy:
	Estimates the joint entropy :math:`h(x_c, x_d)`, where :math:`x_c` is continuous
	random vector and :math:`x_d` is a discrete random vector.

	.. note::

		We use the identities

		.. math::
			h(x, y) &= h(y) + h(x|y) \\

			        &= h(y) + E \\left[ h(x \\vert y=.) \\right]

		that are true when :math:`x` and :math:`y` are either both continuous or both discrete
		to extend the definition of the joint entropy to the case where one is continuous and
		the other discrete.

		Specifically,

		.. math::
			h(x_c, x_d) = h(x_d) + \sum_{j=1}^q \mathbb{P}(x_d=j) h\\left(x_c \\vert x_d=j \\right).

		:math:`h(x_d)` is estimated using :ref:`discrete_entropy <discrete-entropy>`, :math:`\mathbb{P}(x_d=j)` is estimated using frequencies, and
		:math:`h\\left(x_c \\vert x_d=j \\right)` is estimated using :ref:`least_structured_continuous_entropy <least-structured-continuous-entropy>`.


	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous data generating distribution.

	x_d : (n,) np.array
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c.

	Returns
	-------
	h : float
		The entropy of the least structured distribution as evidenced by average pairwise Spearman
		rank correlations.
	"""
	labels = x_d.copy() if len(x_d.shape) == 1 else np.array([','.join([str(_) for _ in row]) for row in x_d])
	categories = list(set(labels))

	n = labels.shape[0]
	probas = np.array([1.*len(labels[labels==cat])/n for cat in categories])
	h = -np.dot(probas, np.log(probas))
	h += np.sum([probas[i] * least_structured_continuous_entropy(x_c[labels==categories[i]]) for i in range(len(categories))])

	return h






#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from kxy.api.core import least_continuous_mutual_information


def information_adjusted_correlation(x, y, p=0):
	"""
	.. _information-adjusted-correlation:
	Calculates the information-adjusted correlation matrix between two arrays.

	.. note::

		Pearson's correlation coefficient quantifies the linear association between two random variables.
		Indeed, two random variables can be statistically dependent despite being decorrelated. Such dependence
		will typically materialize during tail events, the worst timing from a risk management perspective.


		The mutual information rate between two scalar time series provides an alternative that fully captures 
		linear and nonlinear, cross-sectional and temporal dependence:

		.. math::
			I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) = h\\left( \\left\{x_t\\right\} \\right) +  h\\left( \\left\{y_t\\right\} \\right) - h\\left(\\left\{x_t, y_t\\right\}  \\right)

		where :math:`h\\left( \\left\{ x_t \\right\} \\right)` is the entropy rate of the process :math:`\\left\{ x_t \\right\}`. Specifically,
		the mutual information rate is 0 if and only if the two processes are statistically independent, and in particular exhibit no
		cross-sectional or temporal dependence, linear or nonlinear.


		When :math:`\\left\{ x_t, y_t \\right\}` is Gaussian, stationary and memoryless, for instance when :math:`\\left(x_i, y_i \\right)`
		are assumed i.i.d Gaussian, the mutual entropy rate reads

		.. math::
			I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) = -\\frac{1}{2} \log \\left(1- \\text{Corr}\\left(x_t, y_t\\right)^2 \\right).


		We generalize this formula and define the **information-adjusted correlation** as the quantity :math:`\\text{IACorr}\\left( \\left\{x_t\\right\}, \\left\{y_t\\right\} \\right)`
		so that the mutual information rate always reads

		.. math::
			I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) = -\\frac{1}{2} \log \\left(1- \\text{IACorr}\\left( \\left\{x_t\\right\}, \\left\{y_t\\right\} \\right)^2 \\right),

		whether or not the time series are jointly Gaussian and memoryless.

		.. math::
			\\text{IACorr}\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) := \\text{sign}\\left( \\text{Corr}\\left(x_., y_.\\right) \\right)\\sqrt{1-e^{-2 I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right)}}

		where :math:`\\text{sign}(x)=1` if and only if :math:`x \geq 0` and :math:`-1` otherwise. Note that the information-adjusted correlation is 0
		if and only if the two time series are statistically independent, and in particular exhibit no cross-sectional or temporal dependencee.


		Checkout `this blog post <https://medium.com/kxytechnologies/https-medium-com-pit-ai-technologies-the-black-swans-in-your-market-neutral-portfolios-part-1-e17fc18a42a7>`_ 
		for a discussion on the limits of Pearson correlation and beta, and `this other blog post <https://medium.com/kxytechnologies/https-medium-com-pit-ai-technologies-the-black-swans-in-your-market-neutral-portfolios-part-2-b5e8d691b214>`_
		for the introduction of information-adjusted correlation and beta as remedies.

	Parameters
	----------
	x : (n,) or (n, p) np.array
		n i.i.d. draws from a scalar or vector random variable.
	y : (n,) or (n, p) np.array
		n i.i.d. draws from a scalar or vector random variable jointly sampled with x.
	p : int
		The number of lags to use when generating Spearman rank auto-correlation to use 
		as empirical evidence in the maximum-entropy problem. The default value is 0, which 
		corresponds to assuming rows are i.i.d. This is also the only supported value for now.

	Returns
	-------
	c : np.array
		The information-adjusted correlation matrix between the two random variables. 

	Raises
	------
	AssertionError
		If p is different from 0. Higher values will be supported later.
	"""
	assert p==0, 'Only p=0 is supported for now'
	x_ = x[:, None] if len(x.shape) == 1 else x
	y_ = y[:, None] if len(x.shape) == 1 else y
	squared = (x_.shape == y_.shape) and np.allclose(x_, y_)

	c = np.empty((x_.shape[1], y_.shape[1]))
	for i in range(c.shape[0]):
		for j in range(c.shape[1]):
			if squared and i == j:
				c[i, j] = 1.
				continue

			p_corr = np.corrcoef(x_[:, i], y_[:, j])[0, 1]
			s = 1. if p_corr >= 0.0 else -1.
			# TODO: When implemented, replace this with mutual information rate
			mi = least_continuous_mutual_information(x_[:, i], y_[:, j])
			c[i, j] = s * np.sqrt(1.-np.exp(-2.*mi))

	return c

